import { sync as probe_sync } from "probe-image-size";

import { BaseData, BaseView, BlobData } from "@snip/database/types";

import { BaseSnip, BaseSnipArgs, RenderContext, SnipData } from "./base";
import { ImageSnipLegacy } from "./legacy/image";
import { PlaceholderSnip } from "./placeholder";

export interface ImageData extends BaseData {
    blob: BlobData;
    version?: number;
}

export interface ImageView extends BaseView {
    sx: number;
    sy: number;
    swidth: number;
    sheight: number;
    width: number;
    height: number;
}

export interface ImageSnipArgs extends BaseSnipArgs {
    /** The blob data of the image
     * the image width and height are probed from the image
     */
    blob: BlobData;

    /** The x-axis or  y-axis  coordinate of the top left corner of the sub-rectangle of the source
     *  image to draw into the destination context.
     *
     * see https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/drawImage
     *
     * default: 0
     */
    sx?: number;
    sy?: number;

    /** The width/height of the sub-rectangle of the source image to draw into the destination
     * context. If not specified, the entire rectangle from the coordinates specified by
     * rx and ry to the bottom-right corner of the image is used.
     *
     * default: img_width
     */
    swidth?: number;
    sheight?: number;

    /** The desired width/height of the snip in pixels
     *
     * default: img_width
     */
    width?: number;
    height?: number;
}

export class ImageSnip extends BaseSnip {
    public type = "image";
    public version: number = 2;
    public blob: BlobData;

    public bitmap?: CanvasImageSource;

    // Image dimensions in pixels (probed from the image (blob))
    public img_width: number;
    public img_height: number;

    // Size of the sub-rectangle of the source image
    // i.e. the cropped image
    public sx: number;
    public sy: number;
    public swidth: number;
    public sheight: number;

    // Size of the image in the destination context
    // i.e. the rendered image on the canvas
    // base: x,y
    public dwidth: number;
    public dheight: number;

    // Indicate if bitmap was loaded
    public ready: Promise<boolean>;

    // Overlay mode
    public overlay: boolean = false;

    constructor({
        blob,
        sx,
        sy,
        swidth,
        sheight,
        width,
        height,
        ...baseArgs
    }: ImageSnipArgs) {
        super({ ...baseArgs });

        // Set default values
        this.blob = blob;

        // Probe image dimensions
        const prob_res = probe_sync(Buffer.from(blob.data, "base64"));
        if (!prob_res) {
            throw new Error("Could not probe image dimensions");
        }
        this.img_width = prob_res.width;
        this.img_height = prob_res.height;

        // If crop is not defined
        // we use the full image
        this.sx = sx || 0;
        this.sy = sy || 0;
        this.swidth = swidth || this.img_width;
        this.sheight = sheight || this.img_height;

        // Width and height on the canvas
        this.dwidth = width || Math.min(this.img_width, 1400);
        this.dheight =
            height || this.dwidth * (this.img_height / this.img_width);

        // Convert string to base64
        this.ready = this.get_bitmap();
    }

    public get_bitmap() {
        return fetch(`data:${this.blob.mime};base64,${this.blob.data}`)
            .then((response) => response.blob())
            .then((blob) => {
                try {
                    return createImageBitmap(blob);
                } catch (_e) {
                    throw new Error("No ImageBitmap support");
                }
            })
            .then((bitmap) => {
                this.bitmap = bitmap;
                return true;
            })
            .catch((_e) => {
                return false;
            });
    }

    /** Construct a instance of the snip from a
     * SnipData object. This is used to create a snip
     * object from the database data dict.
     *
     * @param {SnipData<ImageData, ImageView>} data - Data dict
     */
    static from_data(
        data: SnipData<ImageData, ImageView>,
    ): ImageSnip | ImageSnipLegacy {
        if (!data.data.blob) {
            throw new Error("Blob is null");
        }
        if (!data.data.blob.mime) {
            throw new Error("Mime is not set!");
        }

        if (!data.data.version) {
            return ImageSnipLegacy.from_data(data);
        }

        return new ImageSnip({
            // Image
            blob: data.data.blob,
            sx: data.view?.sx,
            sy: data.view?.sy,
            swidth: data.view?.swidth,
            sheight: data.view?.sheight,
            width: data.view?.width,
            height: data.view?.height,
            //Base
            x: data.view?.x,
            y: data.view?.y,
            rot: data.view?.rot,
            mirror: data.view?.mirror,
            id: data.id,
            page_id: data.page_id,
            book_id: data.book_id,
            last_updated: data.last_updated,
            created: data.created,
        });
    }

    static from_legacy(data: ImageSnipLegacy): ImageSnip {
        // Parse rot string "90m" -> 90 true

        return new ImageSnip({
            // Image
            blob: data.blob as BlobData,
            rot: data.rot,
            mirror: data.mirror,
            sx: data.sx,
            sy: data.sy,
            swidth: data.swidth,
            sheight: data.sheight,
            width: data.width,
            height: data.height,
            //Base
            x: data.x,
            y: data.y,
            id: data.id,
            page_id: data.page_id,
            book_id: data.book_id,
            last_updated: data.last_updated,
            created: data.created,
        });
    }

    /** Renders the simple text to the canvas function
     *
     * @param {RenderContext} ctx - Canvas context
     */
    render(ctx: RenderContext): void {
        if (!this.bitmap) {
            console.error("No bitmap loaded!");
            return;
        }

        // Apply translate, rotate and mirror
        // see base.ts
        const transform = ctx.getTransform();
        this.transform(ctx);

        // Draw the image in the right scale
        const scale_x = this.width / this.img_width;
        const scale_y = this.height / this.img_height;
        ctx.scale(scale_x, scale_y);
        ctx.drawImage(
            this.bitmap,
            // In image coordinates
            this.sx,
            this.sy,
            this.swidth,
            this.sheight,
            // In canvas coordinates
            this.sx,
            this.sy,
            this.swidth,
            this.sheight,
        );

        // If we are in overlay mode we want to draw the full image too
        if (this.overlay) {
            ctx.globalAlpha = 0.1;
            ctx.scale(1 / scale_x, 1 / scale_y);
            ctx.drawImage(this.bitmap, 0, 0, this.width, this.height);
            ctx.globalAlpha = 1;
        }
        ctx.setTransform(transform);
    }

    /** Create a dictionary in data standard format
     *
     * @returns {Dictionary}
     */
    data(): ImageData {
        return {
            type: "image",
            blob: this.blob,
            version: this.version,
        };
    }
    view(): ImageView {
        const view = super.view() as ImageView;
        view.sx = this.sx;
        view.sy = this.sy;
        view.swidth = this.swidth;
        view.sheight = this.sheight;
        view.width = this.width;
        view.height = this.height;
        return view;
    }

    /** Overload data to allow interaction events */
    public to_data(minimal?: boolean): SnipData<BaseData, BaseView> {
        if (minimal) {
            return PlaceholderSnip.from_snip(this).to_data();
        } else {
            // for some reason the super to_data function does not apply
            // the local data view versions of the img class
            const data = super.to_data();
            data.data = this.data();
            data.view = this.view();
            return data;
        }
    }

    get width(): number {
        return this.dwidth;
    }
    set width(width: number) {
        this.dwidth = width;
    }
    get height(): number {
        return this.dheight;
    }
    set height(height: number) {
        this.dheight = height;
    }
}
